r"""
Degenerate submanifolds

An *embedded (resp. immersed) degenerate submanifold of a proper
pseudo-Riemannian manifold* `(M,g)` is an embedded (resp. immersed)
submanifold `H` of `M` as a differentiable manifold such that pull
back of the metric tensor `g` via the embedding (resp. immersion)
endows `H` with the structure of a degenerate manifold.

Degenerate submanifolds are study in many fields of mathematics and physics,
for instance in Differential Geometry (especially in geometry of
lightlike submanifold) and in General Relativity. In geometry of lightlike
submanifolds, according to the dimension `r` of the radical distribution
(see below for definition of radical distribution), degenerate submanifolds
have been classified into 4 subgroups: `r`-lightlike submanifolds, Coisotropic
submanifolds, Isotropic submanifolds and Totally lightlike submanifolds.
(See the book of Krishan L. Duggal and Aurel Bejancu [DS2010]_.)

In the present module, you can define any of the 4 types but most of the
methods are implemented only for degenerate hypersurfaces who belong to
`r`-lightlike submanifolds. However, they might be generalized to
`1`-lightlike submanifolds. In the literature there is a new approach
(the rigging technique) for studying `1`-lightlike submanifolds but
here we use the method of Krishan L. Duggal and Aurel Bejancu based on
the screen distribution.

Let `H` be a lightlike hypersurface of a pseudo-Riemannian manifold
`(M,g)`. Then the normal bundle `TH^\perp` intersect the tangent
bundle `TH`. The radical distribution is defined as
`Rad(TH)=TH\cap TH^\perp`. In case of hypersurfaces, and more
generally `1`-lightlike submanifolds, this is a rank 1 distribution.
A screen distribution `S(TH)` is a complementary of `Rad(TH)` in `TH`.

Giving a screen distribution `S(TH)` and a null vector field `\xi`
locally defined and spanning `Rad(TH)`, there exists a unique
transversal null vector field 'N' locally defined and such that
`g(N,\xi)=1`. From a transverse vector 'v', the null rigging 'N'
is giving by the formula

.. MATH::

    N = \frac{1}{g(\xi, v)}\left(v-\frac{g(v,v)}{2g(\xi, v)}\xi\right)

Tensors on the ambient manifold `M` are projected on `H` along `N`
to obtain induced objects. For instance, induced connection is the
linear connection defined on H through the Levi-Civitta connection of
`g` along `N`.

To work on a degenerate submanifold, after defining `H` as an instance
of :class:`~sage.manifolds.differentiable.manifold.DifferentiableManifold`,
with the keyword ``structure='degenerate_metric'``, you have to set a
transvervector `v` and screen distribution together with the radical
distribution.

An example of degenerate submanifold from General Relativity is the
horizon of the Schwarzschild black hole. Allow us to recall that
Schwarzschild black hole is the first non-trivial solution of Einstein's
equations. It describes the metric inside a star of radius `R = 2m`,
being `m` the inertial mass of the star. It can be seen as an open
ball in a Lorentzian manifold structure on `\RR^4`::

    sage: M = Manifold(4, 'M', structure='Lorentzian')
    sage: X_M.<t, r, th, ph> = \
    ....: M.chart(r"t r:(0,oo) th:(0,pi):\theta ph:(0,2*pi):\phi")
    sage: var('m'); assume(m>0)
    m
    sage: g = M.metric()
    sage: g[0,0], g[0,1], g[1,1], g[2,2], g[3,3] = \
    ....: -1+2*m/r, 2*m/r, 1+2*m/r, r^2, r^2*sin(th)^2

Let us define the horizon as a degenerate hypersurface::

    sage: H = Manifold(3, 'H', ambient=M, structure='degenerate_metric')
    sage: H
    degenerate hypersurface H embedded in 4-dimensional differentiable
    manifold M

A `2`-dimensional degenerate submanifold of a Lorentzian manifold::

    sage: M = Manifold(4, 'M', structure='Lorentzian')
    sage: X.<t,x,y,z> = M.chart()
    sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
    sage: S
    2-dimensional degenerate submanifold S embedded in 4-dimensional
    differentiable manifold M
    sage: X_S.<u,v> = S.chart()
    sage: Phi = S.diff_map(M, {(X_S, X): [u, u, u, v]},
    ....:         name='Phi', latex_name=r'\Phi');
    sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y]}, name='Phi_inv',
    ....:           latex_name=r'\Phi^{-1}');
    sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
    sage: g = M.metric()
    sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
    sage: S.set_transverse(rigging=[x,y])
    sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
    sage: V = M.vector_field(); V[3] = 1
    sage: Sc = S.screen('Sc', V, xi)

    sage: S.default_screen()
    screen distribution Sc along the 2-dimensional degenerate submanifold
    S embedded in 4-dimensional differentiable manifold M mapped into
    the 4-dimensional Lorentzian manifold M

    sage: S.ambient_metric()
    Lorentzian metric g on the 4-dimensional Lorentzian manifold M

    sage: S.induced_metric()
    degenerate metric gamma on the 2-dimensional degenerate submanifold S
    embedded in 4-dimensional differentiable manifold M

    sage: S.first_fundamental_form()
    Field of symmetric bilinear forms g|S along the 2-dimensional
    degenerate submanifold S embedded in 4-dimensional differentiable manifold M
    with values on the 4-dimensional Lorentzian manifold M

    sage: S.adapted_frame()
    Vector frame (S, (vv_0,vv_1,vv_2,vv_3)) with values on the 4-dimensional Lorentzian manifold M

    sage: S.projection(V)
    Tensor field of type (1,0) along the 2-dimensional degenerate submanifold S
    embedded in 4-dimensional differentiable manifold M
    with values on the 4-dimensional Lorentzian manifold M

    sage: S.weingarten_map()  # long time
    Tensor field nabla_g(xi)|X(S) of type (1,1) along the 2-dimensional
    degenerate submanifold S embedded in 4-dimensional differentiable manifold M
    with values on the 4-dimensional Lorentzian manifold M

    sage: SO = S.shape_operator()  # long time
    sage: SO.display()             # long time
    A^* = 0

    sage: S.is_tangent(xi.along(Phi))
    True
    sage: v = M.vector_field(); v[1] = 1
    sage: S.is_tangent(v.along(Phi))
    False

AUTHORS:

- Hans Fotsing Tetsing (2019) : initial version

REFERENCES:

- [DB1996]_
- [DS2010]_
- [FNO2019]_
"""
# *****************************************************************************
#  Copyright (C) 2019 Hans Fotsing Tetsing <hans.fotsing@aims-cameroon.org>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING

from sage.manifolds.differentiable.degenerate import DegenerateManifold, TangentTensor
from sage.manifolds.differentiable.differentiable_submanifold import (
    DifferentiableSubmanifold,
)
from sage.manifolds.differentiable.pseudo_riemannian import PseudoRiemannianManifold
from sage.manifolds.differentiable.vectorfield_module import VectorFieldModule
from sage.matrix.constructor import matrix
from sage.rings.infinity import infinity
from sage.symbolic.expression import Expression

if TYPE_CHECKING:
    from sage.manifolds.differentiable.metric import DegenerateMetric


class DegenerateSubmanifold(DegenerateManifold, DifferentiableSubmanifold):
    r"""
    Degenerate submanifolds.

    An *embedded (resp. immersed) degenerate submanifold of a proper
    pseudo-Riemannian manifold* `(M,g)` is an embedded (resp. immersed)
    submanifold `H` of `M` as a differentiable manifold such that pull
    back of the metric tensor `g` via the embedding (resp. immersion)
    endows `H` with the structure of a degenerate manifold.

    INPUT:

    - ``n`` -- positive integer; dimension of the manifold
    - ``name`` -- string; name (symbol) given to the manifold
    - ``ambient`` -- (default: ``None``) pseudo-Riemannian manifold `M` in
      which the submanifold is embedded (or immersed). If ``None``, it is set
      to ``self``
    - ``metric_name`` -- (default: ``None``) string; name (symbol) given to the
      metric; if ``None``, ``'g'`` is used
    - ``signature`` -- (default: ``None``) signature `S` of the metric as a
      tuple: `S = (n_+, n_-, n_0)`, where `n_+` (resp. `n_-`, resp. `n_0`) is the
      number of positive terms (resp. negative terms, resp. zero tems) in any
      diagonal writing of the metric components; if ``signature`` is not
      provided, `S` is set to `(ndim-1, 0, 1)`, being `ndim` the manifold's dimension
    - ``base_manifold`` -- (default: ``None``) if not ``None``, must be a
      topological manifold; the created object is then an open subset of
      ``base_manifold``
    - ``diff_degree`` -- (default: ``infinity``) degree of differentiability
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to
      denote the manifold; if none are provided, it is set to ``name``
    - ``metric_latex_name`` -- (default: ``None``) string; LaTeX symbol to
      denote the metric; if none is provided, it is set to ``metric_name``
    - ``start_index`` -- (default: 0) integer; lower value of the range of
      indices used for "indexed objects" on the manifold, e.g., coordinates
      in a chart
      - ``category`` -- (default: ``None``) to specify the category; if
      ``None``, ``Manifolds(field)`` is assumed (see the category
      :class:`~sage.categories.manifolds.Manifolds`)
    - ``unique_tag`` -- (default: ``None``) tag used to force the construction
      of a new object when all the other arguments have been used previously
      (without ``unique_tag``, the
      :class:`~sage.structure.unique_representation.UniqueRepresentation`
      behavior inherited from
      :class:`~sage.manifolds.subset.ManifoldSubset`
      would return the previously constructed object corresponding to these
      arguments)

    .. SEEALSO::

        :mod:`~sage.manifolds.manifold` and
        :mod:`~sage.manifolds.differentiable.differentiable_submanifold`
    """
    def __init__(self, n, name, ambient=None, metric_name=None, signature=None,
                 base_manifold=None, diff_degree=infinity, latex_name=None,
                 metric_latex_name=None, start_index=0, category=None,
                 unique_tag=None):
        r"""
        Construct a degenerate submanifold.

        EXAMPLES:

        A `2`-dimensional degenerate submanifold of a Lorentzian manifold::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: S
            2-dimensional degenerate submanifold S embedded in 4-dimensional
            differentiable manifold M
        """
        DegenerateManifold.__init__(self, n, name=name,
                                          metric_name=metric_name,
                                          signature=signature,
                                          base_manifold=base_manifold,
                                          diff_degree=diff_degree,
                                          latex_name=latex_name,
                                          metric_latex_name=metric_latex_name,
                                          start_index=start_index,
                                          category=category)
        DifferentiableSubmanifold.__init__(self, n, name, self._field,
                                           self._structure, ambient=ambient,
                                           base_manifold=base_manifold,
                                           latex_name=latex_name,
                                           start_index=start_index,
                                           category=category)
        self._normal = None
        self._first_fundamental_form = None
        self._induced_metric = None
        self._second_fundamental_form = {}
        self._ambient_metric = None
        self._gauss_curvature = {}
        self._principal_directions = {}
        self._principal_curvatures = {}
        self._mean_curvature = {}
        self._screens = {}
        self._default_screen = None
        self._adapted_frame = {}
        self._shape_operator = {}
        self._rotation_one_form = {}
        signature = self._ambient.metric().signature()
        ndim = self._ambient._dim
        try:
            if signature[0] == ndim or signature[1] == ndim:
                raise ValueError("ambient must be a proper pseudo-Riemannian"
                                 " or a degenerate manifold")
        except TypeError:
            if signature == ndim or signature == -ndim:
                raise ValueError("ambient must be a proper pseudo-Riemannian"
                                 " or a degenerate manifold")
        self._transverse = {}

    def _repr_(self):
        r"""
        Return a string representation of the submanifold.

        If no ambient manifold is specified, the submanifold is considered
        as a manifold.

        TESTS:

        A `2`-dimensional degenerate submanifold of a Lorentzian manifold::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: S.__repr__()
            '2-dimensional degenerate submanifold S embedded in 4-dimensional
            differentiable manifold M'
        """
        if self._ambient is None:
            return super(DegenerateManifold, self).__repr__()
        if self._ambient._dim - self._dim == 1:
            return "degenerate hypersurface {} embedded " \
                "in {}-dimensional differentiable " \
                "manifold {}".format(self._name, self._ambient._dim,
                                     self._ambient._name)
        return "{}-dimensional degenerate submanifold {} embedded " \
               "in {}-dimensional differentiable " \
               "manifold {}".format(self._dim, self._name, self._ambient._dim,
                                    self._ambient._name)

    def ambient_metric(self):
        r"""
        Return the metric of the ambient manifold. The submanifold has to be
        embedded

        OUTPUT: the metric of the ambient manifold

        EXAMPLES:

        The lightcone of the 3D Minkowski space::

            sage: M = Manifold(3, 'M', structure='Lorentzian')
            sage: X.<t,x,y> = M.chart()
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [sqrt(u^2+v^2), u, v]},
            ....:           name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x, y]}, name='Phi_inv',
            ....:               latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: S.ambient_metric()
            Lorentzian metric g on the 3-dimensional Lorentzian manifold M
        """
        if self._ambient_metric is None:
            if not self._embedded or not isinstance(self._ambient,
                            (PseudoRiemannianManifold, DegenerateManifold)):
                raise ValueError("degenerate submanifold must be embedded in a "
                                 "pseudo-Riemannian or degenerate manifold")
            self._ambient_metric = self._ambient.metric()
        return self._ambient_metric

    def default_screen(self):
        r"""
        Return the default screen distribution.

        OUTPUT:

        - an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi)  # long time
            sage: S.default_screen()              # long time
            screen distribution Sc along the degenerate hypersurface S embedded
            in 4-dimensional differentiable manifold M mapped into the 4-dimensional
            Lorentzian manifold M
        """
        return self._default_screen

    def list_of_screens(self):
        r"""
        Return the default screen distribution.

        OUTPUT:

        - an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:                  name='Phi', latex_name=r'\Phi')
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:                      latex_name=r'\Phi^{-1}')
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi)  # long time
            sage: S.list_of_screens()             # long time
            {'Sc': screen distribution Sc along the degenerate hypersurface S
            embedded in 4-dimensional differentiable manifold M mapped into the
            4-dimensional Lorentzian manifold M}
        """
        return self._screens

    def set_transverse(self, rigging=None, normal=None):
        r"""
        For setting a transversal distribution of the degenerate submanifold.

        According to the type of the submanifold among the 4 possible types,
        one must enter a list of normal transversal vector fields and/or a
        list of transversal and not normal vector fields spanning a transverse
        distribution.

        INPUT:

        - ``rigging`` -- list or tuple (default: ``None``); list of vector fields
          of the ambient manifold or chart function; of the ambient manifold in
          the latter case, the corresponding gradient vector field with respect to
          the ambient metric is calculated; the vectors must be linearly independent,
          transversal to the submanifold but not normal
        - ``normal`` -- list or tuple (default: ``None``); list of vector fields
          of the ambient manifold or chart function; of the ambient manifold in
          the latter case, the corresponding gradient vector field with respect to
          the ambient metric is calculated; the vectors must be linearly independent,
          transversal and normal to the submanifold

        EXAMPLES:

        The lightcone of the 3-dimensional Minkowski space `\RR^3_1`::

            sage: M = Manifold(3, 'M', structure='Lorentzian')
            sage: X.<t,x,y> = M.chart()
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [sqrt(u^2+v^2), u, v]},
            ....:                   name='Phi', latex_name=r'\Phi')
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x, y]}, name='Phi_inv',
            ....:                           latex_name=r'\Phi^{-1}')
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2] = -1,1,1
            sage: S.set_transverse(rigging=t)
        """
        if isinstance(rigging, (list, tuple)):
            rigging = list(rigging)
        else:
            if rigging is not None:
                rigging = [rigging]
        if isinstance(normal, (list, tuple)):
            normal = list(normal)
        else:
            if normal is not None:
                normal = [normal]
        frame = self.default_frame()
        im = self.immersion()
        g = self.ambient_metric().along(im)
        nor = []
        rig = []
        l1, l2 = 0, 0
        if normal is not None:
            for u in normal:
                if isinstance(u, Expression):
                    u = self._ambient.scalar_field(u).gradient()
                for v in frame:
                    v = im.pushforward(v)
                    if not g(u.along(im), v).is_zero():
                        raise ValueError("{} is not tangent to {}".format(u.display(), self._name))
                nor.append(u)
                l1 += 1
        if rigging is not None:
            for u in rigging:
                if isinstance(u, Expression):
                    u = self._ambient.scalar_field(u).gradient()
                rigg = False
                for v in frame:
                    v = im.pushforward(v)
                    if not g(u.along(im), v).is_zero():
                        rigg = True
                if not rigg:
                    raise ValueError("{} is normal to {}".format(u.display(), self._name))
                rig.append(u)
                l2 += 1
        if l1+l2 != self._codim:
            raise ValueError("length of the transverse must be {}".format(self._codim))
        self._transverse['normal'] = tuple(nor)
        self._transverse['rigging'] = tuple(rig)

    def screen(self, name, screen, rad, latex_name=None):
        r"""
        For setting a screen distribution and vector fields of the radical distribution
        that will be used for computations

        INPUT:

        - ``name`` -- string (default: ``None``); name given to the screen
        - ``latex_name`` -- string (default: ``None``); LaTeX symbol to denote
          the screen; if ``None``, the LaTeX symbol is set to ``name``
        - ``screen`` -- list or tuple  of vector fields
          of the ambient manifold or chart function; of the ambient manifold in
          the latter case, the corresponding gradient vector field with respect to
          the ambient metric is calculated; the vectors must be linearly independent,
          tangent to the submanifold but not normal
        - ``rad`` -- -- list or tuple  of vector fields
          of the ambient manifold or chart function; of the ambient manifold in
          the latter case, the corresponding gradient vector field with respect to
          the ambient metric is calculated; the vectors must be linearly independent,
          tangent and normal to the submanifold

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi); Sc  # long time
            screen distribution Sc along the degenerate hypersurface S embedded
            in 4-dimensional differentiable manifold M mapped into the 4-dimensional
            Lorentzian manifold M
        """
        if isinstance(screen, (list, tuple)):
            screen = list(screen)
        else:
            screen = [screen]
        if isinstance(rad, (list, tuple)):
            rad = list(rad)
        else:
            rad = [rad]
        if name in self._screens:
            if list(screen) == self._screens[name]._screen and list(rad) == self._screens[name]._rad:
                return self._screens[name]
            raise ValueError("a different screen distribution with the "
                             "same name had already been set")
        if len(screen)+len(rad) != self._dim:
            raise ValueError("total length screen+rad must be {}".format(self._dim))
        frame = self.default_frame()
        im = self.immersion()
        g = self.ambient_metric().along(im)
        for (i, u) in enumerate(screen):
            if isinstance(u, Expression):
                u = self._ambient.scalar_field(u).gradient()
            screen[i] = u
            sc = False
            for v in frame:
                v = im.pushforward(v)
                if not g(u.along(im), v).is_zero():
                    sc = True
            if not sc:
                raise ValueError("{} cannot belong to a screen distribution".format(u.display()))
        for (i, u) in enumerate(rad):
            if isinstance(u, Expression):
                u = self._ambient.scalar_field(u).gradient()
            rad[i] = u
            for v in frame:
                v = im.pushforward(v)
                if not g(u.along(im), v).is_zero():
                    raise ValueError("{} is not orthogonal to {}".format(u.display(), self._name))
        for u in self._transverse['rigging']:
            for v in screen:
                if not g(u.along(im), v.along(im)).is_zero():
                    raise ValueError("{} is not orthogonal to the rigging {}".format(v.display(), u.display()))
        self._screens[name] = Screen(self, name, tuple(screen), tuple(rad), latex_name=latex_name)
        self._default_screen = self._screens[name]
        return self._screens[name]

    def induced_metric(self) -> DegenerateMetric:
        r"""
        Return the pullback of the ambient metric.

        OUTPUT:

        - induced metric, as an instance of
          :class:`~sage.manifolds.differentiable.metric.DegenerateMetric`

        EXAMPLES:

        Section of the lightcone of the Minkowski space with a hyperplane
        passing through the origin::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [sqrt(u^2+v^2), u, v, 0]},
            ....:             name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x, y]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: h = S.induced_metric(); h  # long time
            degenerate metric gamma on the 2-dimensional degenerate
            submanifold S embedded in 4-dimensional differentiable manifold M
        """
        if self._induced_metric is None or self._induced_metric._components == {}:
            self._induced_metric = self.metric()
            self._induced_metric.set(
                               self.immersion().pullback(self.ambient_metric()))
            self._induced_metric.set_name("gamma", r"\gamma")
        return self._induced_metric

    def first_fundamental_form(self):
        r"""
        Return the restriction of the ambient metric on vector field
        along the submanifold and tangent to it. It is difference from
        induced metric who gives the pullback of the ambient metric
        on the submanifold.

        OUTPUT:

        - the first fundamental form, as an instance of
          :class:`~sage.manifolds.differentiable.degenerate.TangentTensor`

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: h = S.first_fundamental_form()   # long time
        """
        if self._first_fundamental_form is None:
            g = self.ambient_metric()
            h = g.copy()
            h.set_name(g._name+"|"+self._name, g._latex_name+r"|_"+self._latex_name)
            h = TangentTensor(h, self.immersion())
            self._first_fundamental_form = h
        return self._first_fundamental_form

    def _ambient_decomposition(self, screen=None):
        r"""
        Return a list ``[screen, rad, normal, rig]`` where ``screen``
        is a list a vector fields on the ambient manifold `M` spanning
        the giving screen distribution, ``rad`` a list of vector fields
        spanning radical distribution, ``normal`` list of normal transversal
        vector fields, and ``rig`` list of rigging vector fields
        corresponding to the giving screen.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`;
          if ``None``, the default screen is used.

        OUTPUT:

        - a list of 4 lists, this 1st one being independent vector fields
          spanning the screen distribution, the 2nd one independent vector fields
          spanning the radical distribution, the 3rd one independent vector fields
          spanning the transversal normal distribution, the 4th one being a list
          of independent riggings in `Rig(T\Sigma)` according to the decomposition

        .. MATH::

            TM_{|\Sigma}=S(T\Sigma)\oplus_{orth}((Rad(T\Sigma)\oplus_{orth}(
            T\sigma^\perp\cap tr(TM))\oplus Rig(T\Sigma))

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);   # long time
            sage: S._ambient_decomposition()        # long time
            [[Vector field on the 4-dimensional Lorentzian manifold M,
              Vector field on the 4-dimensional Lorentzian manifold M],
             [Vector field on the 4-dimensional Lorentzian manifold M],
             (),
             [Vector field N on the 4-dimensional Lorentzian manifold M]]
        """
        try:
            normal = self._transverse['normal']
        except KeyError:
            raise ValueError("Set first transverse by using set_transverse")
        if screen is None:
            screen = self.default_screen()
        if isinstance(screen, Screen):
            rad = screen._rad
            screen = screen._screen
            rig = self._transverse['rigging']
        else:
            raise ValueError("set first a screen distribution")
        if self._codim == 1:
            xi = rad[0]
            v = rig[0]
            g = self.ambient_metric()
            N = (1/g(xi, v))*(v-(g(v,v)/(2*g(xi, v)))*xi)
            if not self._adapted_frame:
                N.set_name(name='N')
            else:
                n = len(self._adapted_frame)
                N.set_name(name='N'+str(n))
            rig = [N]
        return [screen, rad, normal, rig]

    def _adapted_frame_(self, screen=None):
        r"""
        Return a frame
        `(e_1,\ldots,e_p, \xi_1,\ldots, \xi_r, v_1,\ldots, v_q, N_1, \ldots, N_n)`
        of the ambient manifold, being `e_i` vector fields
        spanning the giving screen distribution, `\xi_i` vector fields spanning
        radical distribution, `v_i` normal transversal vector fields, `N_i`
        rigging vector fields corresponding to the giving screen.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`;
          if ``None`` default screen is used.

        OUTPUT: a frame on the ambient manifold

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: T = S._adapted_frame_();         # long time
        """

        if screen is None:
            screen = self.default_screen()
        if screen is None:
            raise ValueError("set first a screen distribution")
        if screen._name in self._adapted_frame:
            return self._adapted_frame[screen._name]
        decomposition = self._ambient_decomposition(screen)
        sc, rad = decomposition[0], decomposition[1]
        normal, rigging = decomposition[2], decomposition[3]
        A = self._ambient.automorphism_field()
        i = self._ambient._sindex
        for u in sc:
            for j in self._ambient.irange():
                A[j,i] = u[j]
            i += 1
        for u in rad:
            for j in self._ambient.irange():
                A[j,i] = u[j]
            i += 1
        for u in normal:
            for j in self._ambient.irange():
                A[j,i] = u[j]
            i += 1
        for u in rigging:
            for j in self._ambient.irange():
                A[j,i] = u[j]
            i += 1
        f = self._ambient.default_frame()
        GLHPhi = f.along(self.immersion())[0].parent().general_linear_group()
        if not self._adapted_frame:
            e = f.new_frame(A, 'vv')
        else:
            n = len(self._adapted_frame)
            e = f.new_frame(A, 'vv'+str(n))
        self.set_change_of_frame(f.along(self.immersion()), e.along(
                  self.immersion()), GLHPhi(A.along(self.immersion())))
        b = e.dual_basis()
        if self._codim == 1:
            if not self._adapted_frame:
                e[self._dim-self._sindex].set_name('N')
            else:
                n = len(self._adapted_frame)
                e[self._dim-self._sindex].set_name('N'+str(n))
            e[self._dim-self._sindex-1].set_name('xi', latex_name=r'\xi')
            if not self._adapted_frame:
                b[self._dim-self._sindex].set_name('N^b', latex_name=r'N^\flat')
            else:
                b[self._dim-self._sindex].set_name('N'+str(n)+'^b', latex_name=r'N'+str(n)+r'^\flat')
            b[self._dim-self._sindex-1].set_name('xi^b', latex_name=r'\xi^\flat')
        self._adapted_frame[screen._name] = e
        return e

    def adapted_frame(self, screen=None):
        r"""
        Return a frame
        `(e_1,\ldots,e_p, \xi_1,\ldots, \xi_r, v_1,\ldots, v_q, N_1, \ldots, N_n)`
        of the ambient manifold along the submanifold, being `e_i` vector fields
        spanning the giving screen distribution, `\xi_i` vector fields spanning
        radical distribution, `v_i` normal transversal vector fields, `N_i`
        rigging vector fields corresponding to the giving screen.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          if ``None`` default screen is used.

        OUTPUT: a frame on the ambient manifold along the submanifold

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: T = S.adapted_frame(); T         # long time
            Vector frame (S, (vv_0,vv_1,vv_2,vv_3)) with values on the 4-dimensional
            Lorentzian manifold M
        """
        e = self._adapted_frame_(screen).along(self.immersion())
        b = e.dual_basis()
        if self._codim == 1:
            if not self._adapted_frame:
                e[self._dim-self._sindex].set_name('N')
            else:
                n = len(self._adapted_frame)
                e[self._dim-self._sindex].set_name('N'+str(n))
            e[self._dim-self._sindex-1].set_name('xi', latex_name=r'\xi')
            if not self._adapted_frame:
                b[self._dim-self._sindex].set_name('N^b', latex_name=r'N^\flat')
            else:
                b[self._dim-self._sindex].set_name('N'+str(n)+'^b', latex_name=r'N'+str(n)+r'^\flat')
            b[self._dim-self._sindex-1].set_name('xi^b', latex_name=r'\xi^\flat')
        return self._adapted_frame_(screen).along(self.immersion())

    def second_fundamental_form(self, screen=None):
        r"""

        This method is implemented only for null hypersurfaces. The method
        returns a tensor `B` of type `(0,2)` instance of
        :class:`~sage.manifolds.differentiable.degenerate.TangentTensor`
        such that for two vector fields `U, V` on the ambient manifold along
        the null hypersurface, one has:

        .. MATH::

            \nabla_UV=D(U,V)+B(U,V)N

        being `\nabla` the ambient connection, `D` the induced connection
        and `N` the chosen rigging.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          If ``None``, the default screen is used

        OUTPUT:

        - an instance of
          :class:`~sage.manifolds.differentiable.degenerate.TangentTensor`

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);   # long time
            sage: B = S.second_fundamental_form();  # long time
            sage: B.display()                       # long time
            B = 0
        """
        if self._ambient._dim-self._dim != 1:
            raise ValueError("'second_fundamental_form' is defined" +
                                      " only for hypersurfaces.")
        if screen is None:
            screen = self.default_screen()
        if screen._name not in self._second_fundamental_form:
            resu = self._ambient.vector_field_module() \
                .tensor((0, 2), name='B', latex_name='B', sym=[(0, 1)], antisym=[]) \
                .along(self.immersion())
            f = self.adapted_frame(screen)
            rad = self._ambient_decomposition(screen)[1]
            nab = self.ambient_metric().connection()
            xi = rad[0]
            A = self.screen_projection(nab(xi))
            resu[f, :] = A[f, :]
            im = self.immersion()
            self._second_fundamental_form[screen._name] = TangentTensor(resu, im)
        return self._second_fundamental_form[screen._name]

    extrinsic_curvature = second_fundamental_form

    def projection(self, tensor, screen=None):
        r"""

        For a given tensor `T` of type `(r, 1)` on the ambient manifold, this
        method returns the tensor `T'` of type `(r,1)` such that for `r`
        vector fields `v_1,\ldots,v_r`, `T'(v_1,\ldots,v_r)` is the projection
        of  `T(v_1,\ldots,v_r)` on ``self`` along the bundle spanned by the
        transversal vector fields provided by :meth:`set_transverse`.

        INPUT:

        - ``tensor`` -- a tensor of type `(r,1)` on the ambient manifold

        OUTPUT:

        - a tensor of type `(r,1)` on the ambient manifold along ``self``

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: U1 = S.projection(U)             # long time
        """
        if tensor.tensor_type()[0] != 1:
            raise NotImplementedError("``projection`` is implemented only for "
                                      "tensors with 1 as contravariant order")
        return TangentTensor(tensor, self.immersion(), screen)

    def screen_projection(self, tensor, screen=None):
        r"""
        For a given tensor `T` of type `(r, 1)` on the ambient manifold, this
        method returns the tensor `T'` of type `(r,1)` such that for `r`
        vector fields `v_1,\ldots,v_r`, `T'(v_1,\ldots,v_r)` is the projection
        of  `T(v_1,\ldots,v_r)` on the bundle spanned by ``screen`` along the
        bundle spanned by the transversal plus the radical vector fields provided.

        INPUT:

        - ``tensor`` -- a tensor of type `(r,1)` on the ambient manifold

        OUTPUT:

        - a tensor of type `(r,1)` on the ambient manifold

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: U1 = S.screen_projection(U);     # long time
        """
        if tensor.tensor_type()[0] != 1:
            raise NotImplementedError("``projection`` is implemented only for " +
                                      "tensors with 1 as contravariant order")
        frame = self.adapted_frame(screen)
        T = tensor.copy()
        try:
            T = T.along(self.immersion())
        except ValueError:
            pass
        T.display(frame)
        for i in self._ambient.index_generator(T.tensor_rank()):
            if i[0] in range(self._dim-self._sindex-1,self._ambient._dim-self._sindex):
                T[frame, i] = 0
        if tensor._latex_name is None:
            T.set_name(tensor._name)
        else:
            T.set_name("P"+tensor._name, latex_name=r'P'+tensor._latex_name)
        return TangentTensor(T, self.immersion(), screen)

    def weingarten_map(self, screen=None):
        r"""

        This method is implemented only for hypersurfaces.
        *Weigarten map* is the `1`-form `W` defined for a vector field
        `U` tangent to ``self`` by

        .. MATH::

            W(U)=\nabla_U\xi

        being `\nabla` the Levi-Civita connection of the ambient manifold
        and `\xi` the chosen vector field spanning the radical distribution.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          If ``None`` the default screen is used.

        OUTPUT:

        - tensor of type `(1,1)` instance of
          :class:`~sage.manifolds.differentiable.degenerate.TangentTensor`

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: v = M.vector_field(); v[1] = 1
            sage: S.set_transverse(rigging=v)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: W = S.weingarten_map();          # long time
            sage: W.display()                      # long time
            nabla_g(xi)|X(S) = 0
        """

        im = self.immersion()
        rad = self._ambient_decomposition(screen)[1]
        nab = self.ambient_metric().connection()
        xi = rad[0]
        T = self.projection(nab(xi)).extension()
        try:
            T = T.along(im)
        except ValueError:
            pass
        T.set_name("nabla_g(xi)|X("+self._name+")",
                   latex_name=r'\nabla_g(\xi)|_{\mathfrak{X}('+self._latex_name+r')}')
        return TangentTensor(T, im, screen)

    def shape_operator(self, screen=None):
        r"""

        This method is implemented only for hypersurfaces.
        *shape operator* is the projection of the Weingarten map
        on the screen distribution along the radical distribution.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          If ``None`` the default screen is used.

        OUTPUT:

        - tensor of type `(1,1)` instance of
          :class:`~sage.manifolds.differentiable.degenerate.TangentTensor`

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: v = M.vector_field(); v[1] = 1
            sage: S.set_transverse(rigging=v)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: SO = S.shape_operator();         # long time
            sage: SO.display()                     # long time
            A^* = 0
        """
        if screen is None:
            screen = self.default_screen()
        if screen._name in self._shape_operator:
            return self._shape_operator[screen._name]
        im = self.immersion()
        rad = self._ambient_decomposition(screen)[1]
        nab = self.ambient_metric().connection()
        xi = rad[0]
        T = self.screen_projection(nab(-xi), screen=screen).extension()
        try:
            T = T.along(im)
        except ValueError:
            pass
        T.set_name("A^*", latex_name=r'A^\ast')
        A = TangentTensor(T, im)
        self._shape_operator[screen._name] = A
        return A

    def gauss_curvature(self, screen=None):
        r"""
        Gauss curvature is the product of all eigenfunctions of the shape operator.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          If ``None`` the default screen is used.

        OUTPUT: a scalar function on ``self``

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: K = S.gauss_curvature();         # long time
            sage: K.display()                      # long time
            S → ℝ
            (u, v, w) ↦ 0
        """
        if self._ambient._dim-self._dim != 1:
            raise ValueError("'gauss_curvature' is defined"
                             " only for hypersurfaces.")
        if screen is None:
            screen = self.default_screen()
        if screen._name not in self._gauss_curvature:
            f = self.adapted_frame()
            A = self.shape_operator()
            self._gauss_curvature[screen._name] = self.scalar_field(
              {chart: A[f,:,chart].determinant()
                for chart in self.top_charts()})
        return self._gauss_curvature[screen._name]

    def principal_directions(self, screen=None):
        r"""

        Principal directions are eigenvectors of the shape operator. This
        method is implemented only for hypersurfaces.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          If ``None`` default screen is used.

        OUTPUT:

        - list of pairs (vector field, scalar field) representing the
          principal directions and the associated principal curvatures

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi); T = S.adapted_frame()  # long time
            sage: PD = S.principal_directions()                          # long time
            sage: PD[2][0].display(T)                                    # long time
            e_2 = xi
        """
        if self._codim != 1:
            raise ValueError("'principal directions' is defined" +
                                      " only for hypersurfaces.")
        if screen is None:
            screen = self.default_screen()
        if screen._name in self._principal_directions:
            return self._principal_directions[screen._name]
        a = self.shape_operator(screen)
        frame = self.adapted_frame(screen)
        pr_d = matrix(
            [[a[frame, :][i, j].expr() for i in self.irange()]
             for j in self.irange()]).eigenvectors_right()
        res = []
        counter = self.irange()
        for eigen_space in pr_d:
            for eigen_vector in eigen_space[1]:
                v = self._ambient.vector_field(name="e_{}".format(next(counter))
                                                            ).along(self.immersion())
                v[frame, :] = list(eigen_vector) + [0]
                res.append((TangentTensor(v, self.immersion()), self.scalar_field(
                  {chart: eigen_space[0] for chart in self.top_charts()})))
                #res[-1][0].set_name("e_{}".format(next(counter)))
        self._principal_directions[screen._name] = res
        return res

    def mean_curvature(self, screen=None):
        r"""

        Mean curvature is the sum of principal curvatures. This
        method is implemented only for hypersurfaces.

        INPUT:

        - ``screen`` -- (default: ``None``) an instance of
          :class:`~sage.manifolds.differentiable.degenerate_submanifold.Screen`.
          If ``None`` the default screen is used.

        OUTPUT: the mean curvature, as a scalar field on the submanifold

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: S.set_transverse(rigging=x)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: m = S.mean_curvature(); m        # long time
            Scalar field on the degenerate hypersurface S embedded in 4-dimensional
            differentiable manifold M
            sage: m.display()                      # long time
            S → ℝ
            (u, v, w) ↦ 0
        """
        if self._codim != 1:
            raise ValueError("'mean_curvature' is defined" +
                    " only for hypersurfaces.")
        if screen is None:
            screen = self.default_screen()
        if screen._name in self._mean_curvature:
            return self._mean_curvature[screen._name]
        pc = [elt[-1] for elt in self.principal_directions(screen)]
        self._mean_curvature[screen._name] = self.scalar_field({chart: sum(
            pc).expr(chart)/self._dim for chart in self.top_charts()})
        return self._mean_curvature[screen._name]

    def is_tangent(self, v):
        r"""
        Determine whether a vector field on the ambient manifold along ``self``
        is tangent to ``self`` or not.

        INPUT:

        - ``v`` -- field on the ambient manifold along ``self``

        OUTPUT:

        - ``True`` if ``v`` is everywhere tangent to ``self`` or ``False`` if
          not

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: v = M.vector_field(); v[1] = 1
            sage: S.set_transverse(rigging=v)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);  # long time
            sage: S.is_tangent(xi.along(Phi))      # long time
            True
            sage: S.is_tangent(v.along(Phi))       # long time
            False
        """
        g = self.ambient_metric()
        im = self.immersion()
        decomposition = self._ambient_decomposition()
        rad, normal = decomposition[1], decomposition[2]
        for u in rad:
            if not g.along(im)(u.along(im),v).is_zero():
                return False
        return all(g.along(im)(u.along(im), v).is_zero() for u in normal)


#**************************************************************************************

class Screen(VectorFieldModule):
    r"""
    Let `H` be a lightlike submanifold embedded in a pseudo-Riemannian
    manifold `(M,g)` with `\Phi` the embedding map. A screen distribution
    is a complementary `S(TH)` of the radical distribution `Rad(TM)=TH\cap
    TH^\perp` in `TH`. One then has

    .. MATH::

        TH=S(TH)\oplus_{orth}Rad(TH)

    INPUT:

    - ``submanifold`` -- a lightlike submanifold, as an instance of
      :class:`DegenerateSubmanifold`
    - ``name`` -- name given to the screen distribution
    - ``screen`` -- vector fields of the ambient manifold which
      span the screen distribution
    - ``rad`` -- vector fields of the ambient manifold which
      span the radical distribution
    - ``latex_name`` -- (default: ``None``) LaTeX symbol to denote the
      screen distribution; if ``None``, it is formed from ``name``

    EXAMPLES:

    The horizon of the Schwarzschild black hole::

        sage: M = Manifold(4, 'M', structure='Lorentzian')
        sage: X_M.<t, r, th, ph> = \
        ....: M.chart(r"t r:(0,oo) th:(0,pi):\theta ph:(0,2*pi):\phi")
        sage: var('m'); assume(m>0)
        m
        sage: g = M.metric()
        sage: g[0,0], g[0,1], g[1,1], g[2,2], g[3,3] = \
        ....: -1+2*m/r, 2*m/r, 1+2*m/r, r^2, r^2*sin(th)^2
        sage: H = Manifold(3, 'H', ambient=M, structure='degenerate_metric')
        sage: X_H.<ht,hth,hph> = \
        ....: H.chart(r"ht:(-oo,oo):t hth:(0,pi):\theta hph:(0,2*pi):\phi")
        sage: Phi = H.diff_map(M, {(X_H, X_M): [ht, 2*m,hth, hph]}, \
        ....: name='Phi', latex_name=r'\Phi')
        sage: Phi_inv = M.diff_map(H, {(X_M, X_H): [t,th, ph]}, \
        ....: name='Phi_inv', latex_name=r'\Phi^{-1}')
        sage: H.set_immersion(Phi, inverse=Phi_inv); H.declare_embedding()
        sage: xi = M.vector_field(-1, 0, 0, 0)
        sage: v = M.vector_field(r, -r, 0, 0)
        sage: e1 = M.vector_field(0, 0, 1, 0)
        sage: e2 = M.vector_field(0, 0, 0, 1)

    A screen distribution for the Schwarzschild black hole horizon::

        sage: H.set_transverse(rigging=v)
        sage: S = H.screen('S', [e1, e2], (xi)); S  # long time
        screen distribution S along the degenerate hypersurface H embedded
        in 4-dimensional differentiable manifold M mapped into the
        4-dimensional Lorentzian manifold M

    The corresponding normal tangent null vector field and null
    transversal vector field::

        sage: xi = S.normal_tangent_vector(); xi.display()  # long time
        xi = -∂/∂t
        sage: N = S.rigging(); N.display()  # long time
        N = ∂/∂t - ∂/∂r

    Those vector fields are normalized by `g(\xi,N)=1`::

        sage: g.along(Phi)(xi, N).display()  # long time
        g(xi,N): H → ℝ
        (ht, hth, hph) ↦ 1

    """

    def __init__(self, submanifold, name, screen, rad, latex_name=None):
        r"""

        TESTS::

            sage: M = Manifold(3, 'M', structure='Lorentzian')
            sage: X.<t,x,y> = M.chart()
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [sqrt(u^2+v^2), u, v]},
            ....:                  name='Phi', latex_name=r'\Phi')
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x, y]}, name='Phi_inv',
            ....:                      latex_name=r'\Phi^{-1}')
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2] = -1,1,1
            sage: S.set_transverse(rigging=t)
            sage: xi = M.vector_field(); xi[0] = sqrt(x^2+y^2); xi[1] = x; xi[2] = y
            sage: U = M.vector_field(); U[1] = -y; U[2] = x
            sage: Sc = S.screen('Sc', U, xi);
        """
        if not isinstance(submanifold, DegenerateSubmanifold):
            raise TypeError("the first argument must be a null submanifold")
        VectorFieldModule.__init__(self, submanifold, dest_map=submanifold.immersion())
        try:
            self._rigging = submanifold._transverse['rigging']
        except KeyError:
            raise ValueError("set first transverse by using `set_transverse` method")
        self._screen = list(screen)
        self._rad = list(rad)
        self._name = name
        if not latex_name:
            latex_name = name
        self._latex_name = latex_name

    def _repr_(self):
        r"""
        String representation of ``self``.

        TESTS::

            sage: M = Manifold(3, 'M', structure='Lorentzian')
            sage: X.<t,x,y> = M.chart()
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [sqrt(u^2+v^2), u, v]},
            ....:                  name='Phi', latex_name=r'\Phi')
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x, y]}, name='Phi_inv',
            ....:                      latex_name=r'\Phi^{-1}')
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2] = -1,1,1
            sage: S.set_transverse(rigging=t)
            sage: xi = M.vector_field(); xi[0] = sqrt(x^2+y^2); xi[1] = x; xi[2] = y
            sage: U = M.vector_field(); U[1] = -y; U[2] = x
            sage: Sc = S.screen('Sc', U, xi); Sc._repr_()
            'screen distribution Sc along the degenerate hypersurface S embedded in
            3-dimensional differentiable manifold M mapped into the 3-dimensional
            Lorentzian manifold M'
        """
        description = "screen distribution "+self._name
        if self._dest_map is self._domain.identity_map():
            description += " on the {}".format(self._domain)
        else:
            description += (" along the {}".format(self._domain)
                            + " mapped into the {}".format(self._ambient_domain))
        return description

    def __getitem__(self, i):
        r"""
        Access vector fields spanning the screen distribution.

        INPUT:

        - ``i`` -- index of the coordinate; if the slice ``[:]``, then all
          the coordinates are returned

        OUTPUT:

        - a vector field on the ambient manifold that belong to
          the screen distribution

        TESTS::

            sage: M = Manifold(3, 'M', structure='Lorentzian')
            sage: X.<t,x,y> = M.chart()
            sage: S = Manifold(2, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [sqrt(u^2+v^2), u, v]},
            ....:                  name='Phi', latex_name=r'\Phi')
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x, y]}, name='Phi_inv',
            ....:                      latex_name=r'\Phi^{-1}')
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2] = -1,1,1
            sage: S.set_transverse(rigging=t)
            sage: xi = M.vector_field(); xi[0] = sqrt(x^2+y^2); xi[1] = x; xi[2] = y
            sage: U = M.vector_field(); U[1] = -y; U[2] = x
            sage: Sc = S.screen('Sc', U, xi);
            sage: Sc.__getitem__(0)
            Vector field along the degenerate hypersurface S embedded in
            3-dimensional differentiable manifold M with values on the 3-dimensional
            Lorentzian manifold M
        """
        sc = [elt.along(self._domain.immersion()) for elt in self._screen]
        return sc[i-self._domain._sindex]

    def normal_tangent_vector(self):
        r"""
        Return either a list ``Rad`` of vector fields spanning the radical
        distribution or (in case of a hypersurface) a normal tangent null
        vector field spanning the radical distribution.

        OUTPUT:

        - either a list of vector fields or a single vector field in
          case of a hypersurface

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:                  name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:                      latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: v = M.vector_field(); v[1] = 1
            sage: S.set_transverse(rigging=v)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);                  # long time
            sage: Rad = Sc.normal_tangent_vector(); Rad.display()  # long time
            xi = ∂/∂t + ∂/∂x
        """
        rad = [elt.along(self._domain.immersion()) for elt in self._rad]
        if self._domain._codim == 1:
            xi = rad[0]
            xi.set_name(name='xi', latex_name=r'\xi')
            return xi
        return rad

    def rigging(self):
        r"""
        Return either a list ``Rad`` of vector fields spanning the
        complementary of the normal distribution `TH^\perp` in the
        transverse bundle or (when `H` is a null hypersurface) the
        null transversal vector field defined in [DB1996]_.

        OUTPUT:

        - either a list made by vector fields or a vector field in
          case of hypersurface

        EXAMPLES:

        A degenerate hyperplane the 4-dimensional Minkowski space `\RR^4_1`::

            sage: M = Manifold(4, 'M', structure='Lorentzian')
            sage: X.<t,x,y,z> = M.chart()
            sage: S = Manifold(3, 'S', ambient=M, structure='degenerate_metric')
            sage: X_S.<u,v,w> = S.chart()
            sage: Phi = S.diff_map(M, {(X_S, X): [u, u, v, w]},
            ....:         name='Phi', latex_name=r'\Phi');
            sage: Phi_inv = M.diff_map(S, {(X, X_S): [x,y, z]}, name='Phi_inv',
            ....:           latex_name=r'\Phi^{-1}');
            sage: S.set_immersion(Phi, inverse=Phi_inv); S.declare_embedding()
            sage: g = M.metric()
            sage: g[0,0], g[1,1], g[2,2], g[3,3] = -1,1,1,1
            sage: v = M.vector_field(); v[1] = 1
            sage: S.set_transverse(rigging=v)
            sage: xi = M.vector_field(); xi[0] = 1; xi[1] = 1
            sage: U = M.vector_field(); U[2] = 1; V = M.vector_field(); V[3] = 1
            sage: Sc = S.screen('Sc', (U,V), xi);    # long time
            sage: rig = Sc.rigging(); rig.display()  # long time
            N = -1/2 ∂/∂t + 1/2 ∂/∂x
        """
        im = self._domain.immersion()
        rig = [elt.along(im) for elt in self._domain._transverse['rigging']]
        if self._domain._codim != 1:
            return rig
        xi = self.normal_tangent_vector()
        v = rig[0]
        g = self._domain.ambient_metric().along(im)
        N = (1/g(xi, v))*(v-(g(v,v)/(2*g(xi, v)))*xi)
        N.set_name(name='N')
        return N



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
# distutils: extra_compile_args = -DGSL_DISABLE_DEPRECATED
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_coupling.h":

  double  gsl_sf_coupling_3j(int two_ja, int two_jb, int two_jc, int two_ma, int two_mb, int two_mc)

  int  gsl_sf_coupling_3j_e(int two_ja, int two_jb, int two_jc, int two_ma, int two_mb, int two_mc, gsl_sf_result * result)

  double  gsl_sf_coupling_6j(int two_ja, int two_jb, int two_jc, int two_jd, int two_je, int two_jf)

  int  gsl_sf_coupling_6j_e(int two_ja, int two_jb, int two_jc, int two_jd, int two_je, int two_jf, gsl_sf_result * result)

  double  gsl_sf_coupling_9j(int two_ja, int two_jb, int two_jc, int two_jd, int two_je, int two_jf, int two_jg, int two_jh, int two_ji)

  int  gsl_sf_coupling_9j_e(int two_ja, int two_jb, int two_jc, int two_jd, int two_je, int two_jf, int two_jg, int two_jh, int two_ji, gsl_sf_result * result)




# sage.doctest: needs sage.rings.finite_rings
r"""
Drinfeld module morphisms

This module provides the class
:class:`sage.rings.function_fields.drinfeld_module.morphism.DrinfeldModuleMorphism`.

AUTHORS:
- Antoine Leudière (2022-04)
"""

# *****************************************************************************
#        Copyright (C) 2022 Antoine Leudière <antoine.leudiere@inria.fr>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#                   http://www.gnu.org/licenses/
# *****************************************************************************

from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.misc.latex import latex
from sage.categories.morphism import Morphism
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.element import coerce_binop

from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.matrix.constructor import matrix


class DrinfeldModuleMorphism(Morphism, UniqueRepresentation,
                             metaclass=InheritComparisonClasscallMetaclass):
    r"""
    This class represents Drinfeld `\GF{q}[T]`-module morphisms.

    Let `\phi` and `\psi` be two Drinfeld `\GF{q}[T]`-modules over
    a field `K`. A *morphism of Drinfeld modules* `\phi \to \psi` is an
    Ore polynomial `f \in K\{\tau\}` such that `f \phi_a = \psi_a f` for
    every `a \in \GF{q}[T]`. In our case, this is equivalent to `f
    \phi_T = \psi_T f`. An *isogeny* is a nonzero morphism.

    To create a morphism object, the user should never explicitly
    instantiate :class:`DrinfeldModuleMorphism`, but rather call the
    parent homset with the defining Ore polynomial::

        sage: Fq = GF(4)
        sage: A.<T> = Fq[]
        sage: K.<z> = Fq.extension(3)
        sage: phi = DrinfeldModule(A, [z, z^2 + z, z^2 + z])
        sage: tau = phi.ore_variable()
        sage: ore_pol = tau + z^5 + z^3 + z + 1
        sage: psi = phi.velu(ore_pol)
        sage: morphism = Hom(phi, psi)(ore_pol)
        sage: morphism
        Drinfeld Module morphism:
          From: Drinfeld module defined by T |--> (z^2 + z)*τ^2 + (z^2 + z)*τ + z
          To:   Drinfeld module defined by T |--> (z^5 + z^2 + z + 1)*τ^2 + (z^4 + z + 1)*τ + z
          Defn: τ + z^5 + z^3 + z + 1


    The given Ore polynomial must indeed define a morphism::

        sage: morphism = Hom(phi, psi)(1)
        Traceback (most recent call last):
        ...
        ValueError: Ore polynomial does not define a morphism

    One can get basic data on the morphism::

        sage: morphism.domain()
        Drinfeld module defined by T |--> (z^2 + z)*τ^2 + (z^2 + z)*τ + z
        sage: morphism.domain() is phi
        True

        sage: morphism.codomain()
        Drinfeld module defined by T |--> (z^5 + z^2 + z + 1)*τ^2 + (z^4 + z + 1)*τ + z
        sage: morphism.codomain() is psi
        True

    ::

        sage: morphism.ore_polynomial()
        τ + z^5 + z^3 + z + 1
        sage: morphism.ore_polynomial() is ore_pol
        True

    One can check various properties::

        sage: morphism.is_zero()
        False
        sage: morphism.is_isogeny()
        True
        sage: morphism.is_endomorphism()
        False
        sage: morphism.is_isomorphism()
        False

    TESTS::

        sage: morphism.parent() is Hom(phi, psi)
        True
        sage: Hom(phi, psi)(morphism) == morphism
        True

    ::

        sage: phi.frobenius_endomorphism().parent() == End(phi)
        True
        sage: End(phi)(0).parent() == End(phi)
        True

    For the sake of completeness, we explain how the user can directly
    instantiate the class, even though this should never be explicitly
    done::

        sage: from sage.rings.function_field.drinfeld_modules.morphism import DrinfeldModuleMorphism
        sage: DrinfeldModuleMorphism(Hom(phi, psi), ore_pol)
        Drinfeld Module morphism:
          From: Drinfeld module defined by T |--> (z^2 + z)*τ^2 + (z^2 + z)*τ + z
          To:   Drinfeld module defined by T |--> (z^5 + z^2 + z + 1)*τ^2 + (z^4 + z + 1)*τ + z
          Defn: τ + z^5 + z^3 + z + 1
        sage: DrinfeldModuleMorphism(Hom(phi, psi), ore_pol) is morphism
        True
    """
    @staticmethod
    def __classcall_private__(cls, parent, x):
        """
        Create the morphism.

        INPUT:

        - ``cls`` -- the class ``DrinfeldModuleMorphism``

        - ``parent`` -- the Drinfeld module homset

        - ``x`` -- a morphism of Drinfeld modules or an element
          (either an Ore polynomial or an element in the base
          ring) defining it

        TESTS::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: End(phi)(T + 1)
            Endomorphism of Drinfeld module defined by T |--> τ^2 + τ + z6
              Defn: τ^2 + τ + z6 + 1

        ::

            sage: tau = phi.ore_variable()
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism is Hom(phi, psi)(morphism)
            True

        ::

            sage: from sage.rings.function_field.drinfeld_modules.morphism import DrinfeldModuleMorphism
            sage: morphism = DrinfeldModuleMorphism(Sets(), tau + 1)
            Traceback (most recent call last):
            ...
            TypeError: parent should be a DrinfeldModuleHomset
        """
        from sage.rings.function_field.drinfeld_modules.homset import DrinfeldModuleHomset
        if not isinstance(parent, DrinfeldModuleHomset):
            raise TypeError('parent should be a DrinfeldModuleHomset')
        domain = parent.domain()
        codomain = parent.codomain()
        A = domain.category().function_ring()
        # NOTE: it used to be x.parent() is parent, but this was False.
        # DrinfeldModuleHomset inherits Homset, which does NOT inherit
        # UniqueRepresentation
        if x.parent() == parent:
            # x is a DrinfeldModuleMorphism
            ore_pol = x.ore_polynomial()
        elif domain is codomain and A.has_coerce_map_from(x.parent()):
            # x is in the function field; we return the endomorphism phi_x
            ore_pol = domain(A(x))
        else:
            # x is an Ore polynomial
            ore_pol = domain.ore_polring()(x)
        if ore_pol * domain.gen() != codomain.gen() * ore_pol:
            raise ValueError('Ore polynomial does not define a morphism')
        return cls.__classcall__(cls, parent, ore_pol)

    def __init__(self, parent, ore_pol) -> None:
        r"""
        Initialize ``self``.

        INPUT:

        - ``parent`` -- the Drinfeld module homset

        - ``ore_pol`` -- the Ore polynomial that defines the morphism

        TESTS::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism._domain is phi
            True
            sage: morphism._codomain is psi
            True
            sage: morphism._ore_polynomial == tau + z6^5 + z6^2 + 1
            True
        """
        super().__init__(parent)
        self._domain = parent.domain()
        self._codomain = parent.codomain()
        self._ore_polynomial = ore_pol

    def _latex_(self) -> str:
        r"""
        Return a LaTeX representation of the morphism.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: latex(morphism)
            τ + z_{6}^{5} + z_{6}^{2} + 1
        """
        return f'{latex(self._ore_polynomial)}'

    def _repr_(self) -> str:
        r"""
        Return a string representation of the morphism.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> τ^2 + τ + z6
              To:   Drinfeld module defined by T |--> τ^2 + (z6^4 + z6^2 + 1)*τ + z6
              Defn: τ + z6^5 + z6^2 + 1
        """
        if self.is_identity():
            return f'Identity morphism of {self._domain}'
        if self.is_endomorphism():
            return f'Endomorphism of {self._domain}\n' \
                   f'  Defn: {self._ore_polynomial}'
        return f'Drinfeld Module morphism:\n' \
                   f'  From: {self._domain}\n'  \
                   f'  To:   {self._codomain}\n' \
                   f'  Defn: {self._ore_polynomial}'

    def __hash__(self) -> int:
        r"""
        Return a hash of ``self``.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: hash(morphism)  # random
            -4214883752078138009
        """
        return hash((self._domain, self._codomain, self._ore_polynomial))

    def is_zero(self) -> bool:
        r"""
        Return ``True`` whether the morphism is the zero morphism.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism.is_zero()
            False

        ::

            sage: zero_morphism = End(phi)(0)
            sage: zero_morphism.is_zero()
            True
        """
        return self._ore_polynomial.is_zero()

    def is_identity(self) -> bool:
        r"""
        Return ``True`` whether the morphism is the identity morphism.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: morphism = End(phi)(1)
            sage: morphism.is_identity()
            True

        ::

            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism.is_identity()
            False
        """
        return self._ore_polynomial == 1

    def is_isogeny(self) -> bool:
        r"""
        Return ``True`` whether the morphism is an isogeny.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism.is_isogeny()
            True

        ::

            sage: zero_morphism = End(phi)(0)
            sage: zero_morphism.is_isogeny()
            False

        ::

            sage: identity_morphism = End(phi)(1)
            sage: identity_morphism.is_isogeny()
            True

        ::

            sage: frobenius_endomorphism = phi.frobenius_endomorphism()
            sage: frobenius_endomorphism.is_isogeny()
            True
        """
        return not self.is_zero()

    def is_isomorphism(self) -> bool:
        r"""
        Return ``True`` whether the morphism is an isomorphism.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: morphism.is_isomorphism()
            False

        ::

            sage: zero_morphism = End(phi)(0)
            sage: zero_morphism.is_isomorphism()
            False

        ::

            sage: identity_morphism = End(phi)(1)
            sage: identity_morphism.is_isomorphism()
            True

        ::

            sage: frobenius_endomorphism = phi.frobenius_endomorphism()
            sage: frobenius_endomorphism.is_isomorphism()
            False
        """
        return self.is_isogeny() and self._ore_polynomial.degree() == 0

    def ore_polynomial(self):
        r"""
        Return the Ore polynomial that defines the morphism.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: psi = DrinfeldModule(A, [z6, z6^4 + z6^2 + 1, 1])
            sage: tau = phi.ore_variable()
            sage: morphism = Hom(phi, psi)(tau + z6^5 + z6^2 + 1)
            sage: ore_pol = morphism.ore_polynomial()
            sage: ore_pol
            τ + z6^5 + z6^2 + 1

        ::

            sage: ore_pol * phi(T) == psi(T) * ore_pol
            True
        """
        return self._ore_polynomial

    def degree(self):
        r"""
        Return the degree of this isogeny, that is, by definition,
        the `\tau`-degree of its defining Ore polynomial.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z6> = Fq.extension(6)
            sage: phi = DrinfeldModule(A, [z6, 1, 1])
            sage: f = phi.frobenius_endomorphism()
            sage: f
            Endomorphism of Drinfeld module defined by T |--> τ^2 + τ + z6
              Defn: τ^6
            sage: f.degree()
            6
        """
        return self._ore_polynomial.degree()

    @coerce_binop
    def __add__(self, other):
        r"""
        Return the sum of this morphism and ``other``.

        INPUT:

        - ``other`` -- a morphism of Drinfeld modules in the
          same category

        TESTS::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 0, 1, z])
            sage: tau = phi.ore_variable()
            sage: f = phi.hom(tau + 1)
            sage: g = T * f
            sage: f + g  # indirect doctest
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> z*τ^3 + τ^2 + z
              To:   Drinfeld module defined by T |--> (2*z^2 + 4*z + 4)*τ^3 + (3*z^2 + 2*z + 2)*τ^2 + (2*z^2 + 3*z + 4)*τ + z
              Defn: (2*z^2 + 4*z + 4)*τ^4 + (z + 1)*τ^3 + τ^2 + (2*z^2 + 4*z)*τ + z + 1

        We check that coercion from the function ring works::

            sage: F = phi.frobenius_endomorphism()
            sage: F + T
            Endomorphism of Drinfeld module defined by T |--> z*τ^3 + τ^2 + z
              Defn: (z + 1)*τ^3 + τ^2 + z
        """
        return self.parent()(self.ore_polynomial() + other.ore_polynomial())

    def _composition_(self, other, H):
        r"""
        Return the composite of this morphism and ``other``.

        EXAMPLES::

            sage: Fq = GF(2)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 1, z, z^2])
            sage: f = phi.frobenius_endomorphism()
            sage: f
            Endomorphism of Drinfeld module defined by T |--> z^2*τ^3 + z*τ^2 + τ + z
              Defn: τ^3
            sage: f * f  # indirect doctest
            Endomorphism of Drinfeld module defined by T |--> z^2*τ^3 + z*τ^2 + τ + z
              Defn: τ^6
        """
        return H(self.ore_polynomial() * other.ore_polynomial())

    def inverse(self):
        r"""
        Return the inverse of this morphism.

        Only morphisms defined by constant nonzero Ore
        polynomials are invertible.

        EXAMPLES::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 1, z, z^2])
            sage: f = phi.hom(2); f
            Endomorphism of Drinfeld module defined by T |--> z^2*τ^3 + z*τ^2 + τ + z
              Defn: 2
            sage: f.inverse()
            Endomorphism of Drinfeld module defined by T |--> z^2*τ^3 + z*τ^2 + τ + z
              Defn: 3

        Inversion of general isomorphisms between different Drinfeld modules
        also works::

            sage: g = phi.hom(z); g
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> z^2*τ^3 + z*τ^2 + τ + z
              To:   Drinfeld module defined by T |--> z^2*τ^3 + (z^2 + 2*z + 3)*τ^2 + (z^2 + 3*z)*τ + z
              Defn: z
            sage: g.inverse()
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> z^2*τ^3 + (z^2 + 2*z + 3)*τ^2 + (z^2 + 3*z)*τ + z
              To:   Drinfeld module defined by T |--> z^2*τ^3 + z*τ^2 + τ + z
              Defn: 3*z^2 + 4

        When the morphism is not invertible, an error is raised::

            sage: F = phi.frobenius_endomorphism()
            sage: F.inverse()
            Traceback (most recent call last):
            ...
            ZeroDivisionError: this morphism is not invertible
        """
        return self.__invert__()

    def __invert__(self):
        r"""
        Return the inverse of this morphism.

        TESTS::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: coeffs = [z] + [K.random_element() for _ in range(10)]
            sage: phi = DrinfeldModule(A, coeffs)
            sage: a = K.random_element()
            sage: while a.is_zero():
            ....:     a = K.random_element()
            sage: f = phi.hom(a)
            sage: g = ~f
            sage: (f*g).is_identity()
            True
            sage: (g*f).is_identity()
            True
        """
        if not self.is_isomorphism():
            raise ZeroDivisionError("this morphism is not invertible")
        H = self.codomain().Hom(self.domain())
        return H(~(self.ore_polynomial()[0]))

    def right_gcd(self, other):
        r"""
        Return the right gcd of this morphism and ``other``.

        If `u` and `v` are separable isogenies with the same domain,
        their right gcd is the separable isogeny whose kernel is
        `\ker u \cap \ker v`.

        INPUT:

        - ``other`` -- a morphism of Drinfeld modules with the
          same domain

        EXAMPLES::

            sage: Fq = GF(7)
            sage: A.<T> = Fq[]
            sage: F.<z> = Fq.extension(2)
            sage: phi = DrinfeldModule(A, [z, z, z, 1])
            sage: tau = phi.ore_variable()
            sage: u = phi.hom(tau + 3)
            sage: v = phi.hom(tau + 5)
            sage: u.right_gcd(v)
            Identity morphism of Drinfeld module defined by T |--> τ^3 + z*τ^2 + z*τ + z

        We underline that the right gcd of isogenies does not make sense
        when the isogenies have different domains::

            sage: psi = DrinfeldModule(A, [z, z, 1, 1])
            sage: w = psi.hom(tau + 1)
            sage: u.right_gcd(w)
            Traceback (most recent call last):
            ...
            ValueError: the two morphisms must have the same domain

        SEEALSO::

            :meth:`left_lcm`
        """
        if (not isinstance(other, DrinfeldModuleMorphism)
         or other.domain() is not self.domain()):
            raise ValueError("the two morphisms must have the same domain")
        u = self.ore_polynomial().right_gcd(other.ore_polynomial())
        return self.domain().hom(u)

    def left_lcm(self, other):
        r"""
        Return the left lcm of this morphism and ``other``.

        If `u` and `v` are separable isogenies with the same domain,
        their right gcd is the separable isogeny whose kernel is
        `\ker u + \ker v`.

        INPUT:

        - ``other`` -- a morphism of Drinfeld modules with the
          same domain

        EXAMPLES::

            sage: Fq = GF(7)
            sage: A.<T> = Fq[]
            sage: F.<z> = Fq.extension(2)
            sage: phi = DrinfeldModule(A, [z, z, z, 1])
            sage: tau = phi.ore_variable()
            sage: u = phi.hom(tau + 3)
            sage: v = phi.hom(tau + 5)
            sage: u.left_lcm(v)
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> τ^3 + z*τ^2 + z*τ + z
              To:   Drinfeld module defined by T |--> τ^3 + z*τ^2 + (6*z + 1)*τ + z
              Defn: τ^2 + τ + 1

        We underline that the left lcm of isogenies does not make sense
        when the isogenies have different domains::

            sage: psi = DrinfeldModule(A, [z, z, 1, 1])
            sage: w = psi.hom(tau + 1)
            sage: v.left_lcm(w)
            Traceback (most recent call last):
            ...
            ValueError: the two morphisms must have the same domain

        SEEALSO::

            :meth:`right_gcd`
        """
        if (not isinstance(other, DrinfeldModuleMorphism)
         or other.domain() is not self.domain()):
            raise ValueError("the two morphisms must have the same domain")
        u = self._ore_polynomial.left_lcm(other._ore_polynomial)
        return self.domain().hom(u)

    def _motive_matrix(self):
        r"""
        Return the matrix giving the action of this morphism
        on the motives of the underlying Drinfeld modules.

        For internal use. Do not call this method directly.

        EXAMPLES::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 0, 1])
            sage: tau = phi.ore_variable()
            sage: u = tau^2 + (2*z^2 + 3*z + 3)*tau + (2*z + 3)
            sage: f = phi.hom(u)
            sage: f._motive_matrix()
            [                      T + 3 + z                 3 + 3*z + 2*z^2]
            [(1 + z + z^2)*T + 3 + 2*z - z^2               T + 2 - z + 2*z^2]
        """
        phi = self.domain()
        phiT = phi.gen()
        r = phiT.degree()
        K = phi.base_over_constants_field()
        S = phi.ore_polring()
        Frob = S.twisting_morphism()
        KT = PolynomialRing(K, name='T')

        # The first row:
        # we write u = u0 + u1*phiT + u2*phiT^2 + ...
        u = self.ore_polynomial()
        us = []
        while not u.is_zero():
            u, ui = u.right_quo_rem(phiT)
            us.append(ui)
        l = len(us)
        row = [KT([us[i][j] for i in range(l)]) for j in range(r)]
        rows = [row]

        # The next rows:
        # each row is obtained from the previous one by
        # applying the semi-linear transformation f |-> t*f
        inv = ~phiT[r]
        B = inv * phiT
        T = KT.gen()
        for i in range(1, r):
            twist = [c.map_coefficients(Frob) for c in row]
            row = [(inv * T - B[0]) * twist[-1]]
            row += [twist[j - 1] - B[j] * twist[-1] for j in range(1, r)]
            rows.append(row)

        return matrix(KT, rows)

    def norm(self, ideal=True):
        r"""
        Return the norm of this isogeny.

        INPUT:

        - ``ideal`` -- boolean (default: ``True``); if ``True``, return the
          norm as an ideal in the function ring of the Drinfeld modules; if
          ``False``, return the norm as an element in this function ring (only
          relevant for endomorphisms)

        EXAMPLES::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 0, 1, z])
            sage: tau = phi.ore_variable()
            sage: f = phi.hom(tau + 1)
            sage: f.norm()
            Principal ideal (T + 4) of Univariate Polynomial Ring in T over Finite Field of size 5

        The norm of the Frobenius endomorphism is equal to the characteristic::

            sage: F = phi.frobenius_endomorphism()
            sage: F.norm()
            Principal ideal (T^3 + 3*T + 3) of Univariate Polynomial Ring in T over Finite Field of size 5
            sage: phi.characteristic()
            T^3 + 3*T + 3

        For `a` in the underlying function ring, the norm of the
        endomorphism given by `\phi_a` is `a^r` where `r` is the rank::

            sage: g = phi.hom(T)
            sage: g.norm()
            Principal ideal (T^3) of Univariate Polynomial Ring in T over Finite Field of size 5

            sage: h = phi.hom(T+1)
            sage: h.norm()
            Principal ideal (T^3 + 3*T^2 + 3*T + 1) of Univariate Polynomial Ring in T over Finite Field of size 5

        For endomorphisms, the norm is not an ideal of `A` but it makes
        sense as an actual element of `A`. We can get this element by passing
        in the argument ``ideal=False``::

            sage: phi.hom(2*T).norm(ideal=False)
            3*T^3

            sage: f.norm(ideal=False)
            Traceback (most recent call last):
            ...
            ValueError: norm is defined as an actual element only for endomorphisms
        """
        nu = self._motive_matrix().det()
        # We cast to A
        A = self.domain().function_ring()
        if ideal:
            nu = A([c.in_base() for c in nu.monic().list()])
            return A.ideal(nu)
        if self.domain() is self.codomain():
            return A([c.in_base() for c in nu.list()])
        raise ValueError("norm is defined as an actual element only for endomorphisms")

    def dual_isogeny(self):
        r"""
        Return a dual isogeny to this morphism.

        By definition, a dual isogeny of `f : \phi \to \psi` is an
        isogeny `g : \psi \to \phi` such that the composite `g \circ f`
        is the multiplication by a generator of the norm of `f`.

        EXAMPLES::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 0, 1, z])
            sage: tau = phi.ore_variable()
            sage: f = phi.hom(tau + 1)
            sage: f
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> z*τ^3 + τ^2 + z
              To:   Drinfeld module defined by T |--> (2*z^2 + 4*z + 4)*τ^3 + (3*z^2 + 2*z + 2)*τ^2 + (2*z^2 + 3*z + 4)*τ + z
              Defn: τ + 1
            sage: g = f.dual_isogeny()
            sage: g
            Drinfeld Module morphism:
              From: Drinfeld module defined by T |--> (2*z^2 + 4*z + 4)*τ^3 + (3*z^2 + 2*z + 2)*τ^2 + (2*z^2 + 3*z + 4)*τ + z
              To:   Drinfeld module defined by T |--> z*τ^3 + τ^2 + z
              Defn: z*τ^2 + (4*z + 1)*τ + z + 4

        We check that `f \circ g` (resp. `g \circ f`) is the multiplication
        by the norm of `f`::

            sage: a = f.norm().gen(); a
            T + 4
            sage: g * f == phi.hom(a)
            True

            sage: psi = f.codomain()
            sage: f * g == psi.hom(a)
            True

        TESTS::

            sage: zero = phi.hom(0)
            sage: zero.dual_isogeny()
            Traceback (most recent call last):
            ...
            ValueError: the dual isogeny of the zero morphism is not defined
        """
        if not self.is_isogeny():
            raise ValueError("the dual isogeny of the zero morphism is not defined")
        nu = self._motive_matrix().det().monic()
        A = self.domain().function_ring()
        nu = A([c.in_base() for c in nu.list()])
        dual = self.domain()(nu) // self.ore_polynomial()
        return self.codomain().hom(dual, codomain=self.domain())

    def characteristic_polynomial(self, var='X'):
        r"""
        Return the characteristic polynomial of this endomorphism.

        INPUT:

        - ``var`` -- string (default: ``X``), the name of the
          variable of the characteristic polynomial

        EXAMPLES::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 0, 1, z])

            sage: f = phi.frobenius_endomorphism()
            sage: f.characteristic_polynomial()
            X^3 + (T + 1)*X^2 + (2*T + 3)*X + 2*T^3 + T + 1

        We verify, on an example, that the characteristic polynomial
        of a morphism corresponding to `\phi_a` is `(X-a)^r` where `r`
        is the rank::

            sage: g = phi.hom(T^2 + 1)
            sage: chi = g.characteristic_polynomial()
            sage: chi.factor()
            (X + 4*T^2 + 4)^3

        An example with another variable name::

            sage: f.characteristic_polynomial(var='Y')
            Y^3 + (T + 1)*Y^2 + (2*T + 3)*Y + 2*T^3 + T + 1

        TESTS::

            sage: tau = phi.ore_variable()
            sage: isog = phi.hom(tau + 1)
            sage: isog.characteristic_polynomial()
            Traceback (most recent call last):
            ...
            ValueError: characteristic polynomial is only defined for endomorphisms
        """
        if self.domain() is not self.codomain():
            raise ValueError("characteristic polynomial is only defined for endomorphisms")
        P = self._motive_matrix().charpoly()
        # We cast to the correct parent
        A = self.domain().function_ring()
        parent = PolynomialRing(A, name=var)
        return parent([A([c.in_base() for c in co.list()]) for co in P.list()])

    def charpoly(self, var='X'):
        r"""
        Return the characteristic polynomial of this endomorphism.

        INPUT:

        - ``var`` -- string (default: ``'X'``); the name of the
          variable of the characteristic polynomial

        EXAMPLES::

            sage: Fq = GF(5)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(3)
            sage: phi = DrinfeldModule(A, [z, 0, 1, z])

            sage: f = phi.frobenius_endomorphism()
            sage: chi = f.charpoly()
            sage: chi
            X^3 + (T + 1)*X^2 + (2*T + 3)*X + 2*T^3 + T + 1

        We check that the characteristic polynomial annihilates the
        morphism (Cayley-Hamilton's theorem)::

            sage: chi(f)
            Endomorphism of Drinfeld module defined by T |--> z*τ^3 + τ^2 + z
              Defn: 0

        We verify, on an example, that the characteristic polynomial
        of the morphism corresponding to `\phi_a` is `(X-a)^r` where `r`
        is the rank::

            sage: g = phi.hom(T^2 + 1)
            sage: g.charpoly().factor()
            (X + 4*T^2 + 4)^3

        An example with another variable name::

            sage: f.charpoly(var='Y')
            Y^3 + (T + 1)*Y^2 + (2*T + 3)*Y + 2*T^3 + T + 1
        """
        return self.characteristic_polynomial(var)



from sage.rings.morphism cimport RingHomomorphism
from sage.structure.element cimport Element


cdef class FrobeniusEndomorphism_padics(RingHomomorphism):
    cdef long _degree
    cdef long _power
    cdef long _order

    cpdef Element _call_(self, x)



r"""
Subrings of the Symbolic Ring

Subrings of the symbolic ring can be created via the
:meth:`~sage.symbolic.ring.SymbolicRing.subring` method of
``SR``. This will call :class:`SymbolicSubring <SymbolicSubringFactory>`
of this module.

The following kinds of subrings are supported:

- A symbolic subring of expressions, whose variables are contained in
  a given set of symbolic variables (see
  :class:`SymbolicSubringAcceptingVars`). E.g.
  ::

      sage: SR.subring(accepting_variables=('a', 'b'))
      Symbolic Subring accepting the variables a, b

- A symbolic subring of expressions, whose variables are disjoint to a
  given set of symbolic variables (see
  :class:`SymbolicSubringRejectingVars`). E.g.
  ::

      sage: SR.subring(rejecting_variables=('r', 's'))
      Symbolic Subring rejecting the variables r, s

- The subring of symbolic constants (see
  :class:`SymbolicConstantsSubring`). E.g.
  ::

      sage: SR.subring(no_variables=True)
      Symbolic Constants Subring


TESTS:

In the following we have a couple of tests to see whether the coercion
framework works properly::

    sage: from sage.symbolic.subring import SymbolicSubring
    sage: V = var('a, r, x')
    sage: A = SymbolicSubring(accepting_variables=(a,)); A
    Symbolic Subring accepting the variable a
    sage: R = SymbolicSubring(rejecting_variables=(r,)); R
    Symbolic Subring rejecting the variable r
    sage: C = SymbolicSubring(no_variables=True); C
    Symbolic Constants Subring

::

    sage: sage.categories.pushout.pushout(A, R)
    Symbolic Subring rejecting the variable r
    sage: sage.categories.pushout.pushout(R, C)
    Symbolic Subring rejecting the variable r
    sage: sage.categories.pushout.pushout(C, A)
    Symbolic Subring accepting the variable a
    sage: sage.categories.pushout.pushout(A, SR)
    Symbolic Ring
    sage: sage.categories.pushout.pushout(R, SR)
    Symbolic Ring
    sage: sage.categories.pushout.pushout(C, SR)
    Symbolic Ring

::

    sage: cm = sage.structure.element.get_coercion_model()
    sage: cm.common_parent(A, R)
    Symbolic Subring rejecting the variable r
    sage: cm.common_parent(R, C)
    Symbolic Subring rejecting the variable r
    sage: cm.common_parent(C, A)
    Symbolic Subring accepting the variable a
    sage: cm.common_parent(A, SR)
    Symbolic Ring
    sage: cm.common_parent(R, SR)
    Symbolic Ring
    sage: cm.common_parent(C, SR)
    Symbolic Ring


AUTHORS:

- Daniel Krenn (2015)


Classes and Methods
===================
"""

# ****************************************************************************
# Copyright (C) 2015 Daniel Krenn <dev@danielkrenn.at>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.rings.abc
from sage.categories.pushout import ConstructionFunctor
from sage.structure.factory import UniqueFactory
from sage.symbolic.ring import SR, SymbolicRing


class SymbolicSubringFactory(UniqueFactory):
    r"""
    A factory creating a symbolic subring.

    INPUT:

    Specify one of the following keywords to create a subring.

    - ``accepting_variables`` -- (default: ``None``) a tuple or other
      iterable of variables. If specified, then a symbolic subring of
      expressions in only these variables is created.

    - ``rejecting_variables`` -- (default: ``None``) a tuple or other
      iterable of variables. If specified, then a symbolic subring of
      expressions in variables distinct to these variables is
      created.

    - ``no_variables`` -- boolean (default: ``False``); if set,
      then a symbolic subring of constant expressions (i.e.,
      expressions without a variable) is created.

    EXAMPLES::

        sage: from sage.symbolic.subring import SymbolicSubring
        sage: V = var('a, b, c, r, s, t, x, y, z')

    ::

        sage: A = SymbolicSubring(accepting_variables=(a, b, c)); A
        Symbolic Subring accepting the variables a, b, c
        sage: tuple((v, v in A) for v in V)
        ((a, True), (b, True), (c, True),
         (r, False), (s, False), (t, False),
         (x, False), (y, False), (z, False))

    ::

        sage: R = SymbolicSubring(rejecting_variables=(r, s, t)); R
        Symbolic Subring rejecting the variables r, s, t
        sage: tuple((v, v in R) for v in V)
        ((a, True), (b, True), (c, True),
         (r, False), (s, False), (t, False),
         (x, True), (y, True), (z, True))

    ::

        sage: C = SymbolicSubring(no_variables=True); C
        Symbolic Constants Subring
        sage: tuple((v, v in C) for v in V)
        ((a, False), (b, False), (c, False),
         (r, False), (s, False), (t, False),
         (x, False), (y, False), (z, False))

    TESTS::

        sage: SymbolicSubring(accepting_variables=tuple()) is C
        True

    ::

        sage: SymbolicSubring(rejecting_variables=tuple()) is SR
        True
    """
    def create_key_and_extra_args(
            self, accepting_variables=None, rejecting_variables=None,
            no_variables=False, **kwds):
        r"""
        Given the arguments and keyword, create a key that uniquely
        determines this object.

        See :class:`SymbolicSubringFactory` for details.

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring.create_key_and_extra_args()
            Traceback (most recent call last):
            ...
            ValueError: cannot create a symbolic subring since nothing is specified
            sage: SymbolicSubring.create_key_and_extra_args(
            ....:     accepting_variables=('a',), rejecting_variables=('r',))
            Traceback (most recent call last):
            ...
            ValueError: cannot create a symbolic subring since input is ambiguous
            sage: SymbolicSubring.create_key_and_extra_args(
            ....:     accepting_variables=('a',), no_variables=True)
            Traceback (most recent call last):
            ...
            ValueError: cannot create a symbolic subring since input is ambiguous
            sage: SymbolicSubring.create_key_and_extra_args(
            ....:     rejecting_variables=('r',), no_variables=True)
            Traceback (most recent call last):
            ...
            ValueError: cannot create a symbolic subring since input is ambiguous
        """
        if accepting_variables is None and \
           rejecting_variables is None and \
           not no_variables:
            raise ValueError('cannot create a symbolic subring '
                             'since nothing is specified')
        if accepting_variables is not None and rejecting_variables is not None or \
           rejecting_variables is not None and no_variables or \
           no_variables and accepting_variables is not None:
            raise ValueError('cannot create a symbolic subring '
                             'since input is ambiguous')

        if accepting_variables is not None:
            vars = tuple(accepting_variables)
            if vars:
                cls = SymbolicSubringAcceptingVars
            else:
                cls = SymbolicConstantsSubring
        elif rejecting_variables is not None:
            vars = tuple(rejecting_variables)
            cls = SymbolicSubringRejectingVars
        elif no_variables:
            vars = tuple()
            cls = SymbolicConstantsSubring

        vars = tuple(sorted(iter(SR(v) for v in vars), key=str))
        return (cls, vars), kwds

    def create_object(self, version, key, **kwds):
        r"""
        Create an object from the given arguments.

        See :class:`SymbolicSubringFactory` for details.

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(rejecting_variables=tuple()) is SR  # indirect doctest
            True
        """
        cls, vars = key
        if cls is SymbolicSubringRejectingVars and not vars:
            return SR
        return cls(vars, **kwds)


SymbolicSubring = SymbolicSubringFactory("SymbolicSubring")


class GenericSymbolicSubring(SymbolicRing):

    def __init__(self, vars):
        r"""
        An abstract base class for a symbolic subring.

        INPUT:

        - ``vars`` -- tuple of symbolic variables

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=('a',))  # indirect doctest
            Symbolic Subring accepting the variable a
            sage: SymbolicSubring(rejecting_variables=('r',))  # indirect doctest
            Symbolic Subring rejecting the variable r
            sage: SymbolicSubring(no_variables=True)  # indirect doctest
            Symbolic Constants Subring
            sage: SymbolicSubring(rejecting_variables=tuple())  # indirect doctest
            Symbolic Ring

        ::

            sage: SR.subring(accepting_variables=(0, pi, sqrt(2), 'zzz', I))
            Traceback (most recent call last):
            ...
            ValueError: Invalid variables: 0, I, pi, sqrt(2)
        """
        super().__init__()
        self._vars_ = set(vars)
        if not all(v.is_symbol() for v in self._vars_):
            raise ValueError('Invalid variables: {}'.format(
                ', '.join(str(v) for v in sorted(self._vars_, key=str)
                          if not v.is_symbol())))

    def _repr_variables_(self):
        r"""
        Return a representation string of the variables.

        OUTPUT: string

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=tuple())._repr_variables_()
            'no variable'
            sage: SymbolicSubring(accepting_variables=('a',))._repr_variables_()
            'the variable a'
            sage: SymbolicSubring(accepting_variables=('a', 'b'))._repr_variables_()
            'the variables a, b'
        """
        if not self._vars_:
            s = 'no variable'
        elif len(self._vars_) == 1:
            s = 'the variable '
        else:
            s = 'the variables '
        return s + ', '.join(str(v) for v in sorted(self._vars_, key=str))

    def has_valid_variable(self, variable):
        r"""
        Return whether the given ``variable`` is valid in this subring.

        INPUT:

        - ``variable`` -- a symbolic variable

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.symbolic.subring import GenericSymbolicSubring
            sage: GenericSymbolicSubring(vars=tuple()).has_valid_variable(x)
            Traceback (most recent call last):
            ...
            NotImplementedError: Not implemented in this abstract base class
        """
        raise NotImplementedError('Not implemented in this abstract base class')

    def _element_constructor_(self, x):
        r"""
        Create the element of this subring specified by the input ``x``.

        INPUT:

        - ``x`` -- an object

        OUTPUT: an element of this symbolic subring

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: S = SymbolicSubring(accepting_variables=('a',))
            sage: S('a')  # indirect doctest
            a
            sage: _.parent()
            Symbolic Subring accepting the variable a
            sage: S('x')  # indirect doctest
            Traceback (most recent call last):
            ...
            TypeError: x is not contained in Symbolic Subring accepting the variable a
        """
        expression = super()._element_constructor_(x)
        assert expression.parent() is self
        if not all(self.has_valid_variable(var)
                   for var in expression.variables()):
            raise TypeError('%s is not contained in %s' % (x, self))
        return expression

    def _coerce_map_from_(self, P):
        r"""
        Return whether ``P`` coerces into this symbolic subring.

        INPUT:

        - ``P`` -- a parent

        OUTPUT: boolean or ``None``

        TESTS::

            sage: from sage.symbolic.subring import GenericSymbolicSubring
            sage: GenericSymbolicSubring(vars=tuple()).has_coerce_map_from(SR)  # indirect doctest  # not tested see #19231
            False

        ::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: C = SymbolicSubring(no_variables=True)
            sage: C.has_coerce_map_from(ZZ)  # indirect doctest
            True
            sage: C.has_coerce_map_from(QQ)  # indirect doctest
            True
            sage: C.has_coerce_map_from(RR)  # indirect doctest
            True
            sage: C.has_coerce_map_from(RIF)  # indirect doctest
            True
            sage: C.has_coerce_map_from(CC)  # indirect doctest
            True
            sage: C.has_coerce_map_from(CIF)  # indirect doctest
            True
            sage: C.has_coerce_map_from(AA)  # indirect doctest
            True
            sage: C.has_coerce_map_from(QQbar)  # indirect doctest
            True
            sage: C.has_coerce_map_from(SR)  # indirect doctest
            False
        """
        from sage.rings.infinity import InfinityRing
        from sage.rings.qqbar import AA, QQbar
        from sage.rings.real_lazy import CLF, RLF

        if isinstance(P, type):
            return SR._coerce_map_from_(P)

        if RLF.has_coerce_map_from(P) or \
           CLF.has_coerce_map_from(P) or \
           AA.has_coerce_map_from(P) or \
           QQbar.has_coerce_map_from(P):
            return True

        if (P is InfinityRing or
            isinstance(P, (sage.rings.abc.RealIntervalField,
                           sage.rings.abc.ComplexIntervalField))):
            return True

        if P._is_numerical():
            return P not in (RLF, CLF, AA, QQbar)

    def __eq__(self, other):
        """
        Compare two symbolic subrings.

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: A = SymbolicSubring(accepting_variables=('a',))
            sage: B = SymbolicSubring(accepting_variables=('b',))
            sage: AB = SymbolicSubring(accepting_variables=('a', 'b'))
            sage: A == A
            True
            sage: A == B
            False
            sage: A == AB
            False
        """
        if not isinstance(other, GenericSymbolicSubring):
            return False
        return self._vars_ == other._vars_

    def __ne__(self, other):
        """
        Check whether ``self`` and ``other`` are not equal.

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: A = SymbolicSubring(accepting_variables=('a',))
            sage: B = SymbolicSubring(accepting_variables=('b',))
            sage: AB = SymbolicSubring(accepting_variables=('a', 'b'))
            sage: A != A
            False
            sage: A != B
            True
            sage: A != AB
            True
        """
        return not self == other

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: A = SymbolicSubring(accepting_variables=('a',))
            sage: B = SymbolicSubring(accepting_variables=('b',))
            sage: hash(A) == hash(A)
            True
            sage: hash(A) == hash(B)
            False
        """
        return hash(tuple(sorted(self._vars_)))


class GenericSymbolicSubringFunctor(ConstructionFunctor):
    r"""
    A base class for the functors constructing symbolic subrings.

    INPUT:

    - ``vars`` -- tuple, set, or other iterable of symbolic variables

    EXAMPLES::

        sage: from sage.symbolic.subring import SymbolicSubring
        sage: SymbolicSubring(no_variables=True).construction()[0]  # indirect doctest
        Subring<accepting no variable>

    .. SEEALSO::

        :class:`sage.categories.pushout.ConstructionFunctor`.
    """

    _functor_name = 'GenericSymbolicSubringFunctor'

    rank = 11

    # The symbolic subring construction returns an object admitting a
    # coercion map into the original, not vice versa.
    coercion_reversed = True

    _repr_type_ = 'generic'

    def __init__(self, vars):
        r"""
        See :class:`GenericSymbolicSubringFunctor` for details.

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=('a',)).construction()[0]  # indirect doctest
            Subring<accepting a>
        """
        self.vars = set(vars)
        from sage.categories.rings import Rings
        super().__init__(Rings(), Rings())

    def _repr_variables_(self):
        r"""
        Return a representation string of the variables.

        OUTPUT: string

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F = SymbolicSubring(accepting_variables=('a',)).construction()[0]
            sage: F._repr_variables_()
            'a'
        """
        return ', '.join(str(v) for v in sorted(self.vars, key=str))

    def _repr_(self):
        r"""
        Return a representation string of this functor.

        OUTPUT: string

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=('a',))  # indirect doctest
            Symbolic Subring accepting the variable a
            sage: SymbolicSubring(rejecting_variables=('r',))  # indirect doctest
            Symbolic Subring rejecting the variable r
            sage: SymbolicSubring(no_variables=True)  # indirect doctest
            Symbolic Constants Subring
        """
        return 'Subring<%s%s%s>' % (
            self._repr_type_, ' ' if self._repr_type_ else '',
            self._repr_variables_() if self.vars else 'no variable')

    def merge(self, other):
        r"""
        Merge this functor with ``other`` if possible.

        INPUT:

        - ``other`` -- a functor

        OUTPUT: a functor or ``None``

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F = SymbolicSubring(accepting_variables=('a',)).construction()[0]
            sage: F.merge(F) is F
            True
        """
        if self == other:
            return self

    def __eq__(self, other):
        r"""
        Return whether this functor is equal to ``other``.

        INPUT:

        - ``other`` -- a functor

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F = SymbolicSubring(accepting_variables=('a',)).construction()[0]
            sage: F == F
            True
        """
        return type(self) is type(other) and self.vars == other.vars

    def __ne__(self, other):
        r"""
        Return whether this functor is not equal to ``other``.

        INPUT:

        - ``other`` -- a functor

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F = SymbolicSubring(accepting_variables=('a',)).construction()[0]
            sage: F != F
            False
        """
        return not self == other


class SymbolicSubringAcceptingVars(GenericSymbolicSubring):
    r"""
    The symbolic subring consisting of symbolic expressions in the given variables.
    """

    def _repr_(self):
        r"""
        Return a representation string of this symbolic subring.

        OUTPUT: string

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=('a',))  # indirect doctest
            Symbolic Subring accepting the variable a
        """
        return 'Symbolic Subring accepting %s' % \
            (self._repr_variables_())

    def has_valid_variable(self, variable):
        r"""
        Return whether the given ``variable`` is valid in this subring.

        INPUT:

        - ``variable`` -- a symbolic variable

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: S = SymbolicSubring(accepting_variables=('a',))
            sage: S.has_valid_variable('a')
            True
            sage: S.has_valid_variable('r')
            False
            sage: S.has_valid_variable('x')
            False
        """
        return SR(variable) in self._vars_

    def construction(self):
        r"""
        Return the functorial construction of this symbolic subring.

        OUTPUT:

        A tuple whose first entry is a construction functor and its second
        is the symbolic ring.

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=('a',)).construction()
            (Subring<accepting a>, Symbolic Ring)
        """
        return (SymbolicSubringAcceptingVarsFunctor(self._vars_), SR)

    def _coerce_map_from_(self, P):
        r"""
        Return whether ``P`` coerces into this symbolic subring.

        INPUT:

        - ``P`` -- a parent

        OUTPUT: boolean or ``None``

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: A = SymbolicSubring(accepting_variables=('a',))
            sage: AB = SymbolicSubring(accepting_variables=('a', 'b'))
            sage: A.has_coerce_map_from(AB)  # indirect doctest
            False
            sage: AB.has_coerce_map_from(A)  # indirect doctest
            True
        """
        if isinstance(P, SymbolicSubringAcceptingVars):
            return self._vars_ >= P._vars_
        return super()._coerce_map_from_(P)

    def _an_element_(self):
        r"""
        Return an element of this symbolic subring.

        OUTPUT: a symbolic expression

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(accepting_variables=('a',)).an_element()
            a
            sage: _.parent()
            Symbolic Subring accepting the variable a
        """
        return self(sorted(self._vars_, key=str)[0])


class SymbolicSubringAcceptingVarsFunctor(GenericSymbolicSubringFunctor):

    _functor_name = 'SymbolicSubringAcceptingVarsFunctor'

    _repr_type_ = 'accepting'

    def merge(self, other):
        r"""
        Merge this functor with ``other`` if possible.

        INPUT:

        - ``other`` -- a functor

        OUTPUT: a functor or ``None``

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F = SymbolicSubring(accepting_variables=('a',)).construction()[0]
            sage: G = SymbolicSubring(rejecting_variables=('r',)).construction()[0]
            sage: F.merge(F) is F
            True
            sage: F.merge(G) is G
            True
        """
        if self == other:
            return self
        if type(self) is type(other):
            return type(self)(self.vars | other.vars)
        if isinstance(other, SymbolicSubringRejectingVarsFunctor):
            if not (self.vars & other.vars):
                return other

    def _apply_functor(self, R):
        """
        Apply this functor to the given symbolic ring `R`.

        INPUT:

        - ``R`` -- a symbolic ring

        OUTPUT: a subring of ``R``

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F, R = SymbolicSubring(accepting_variables=('a',)).construction()
            sage: F(R)  # indirect doctest
            Symbolic Subring accepting the variable a

        TESTS::

            sage: F(F(R))
            Traceback (most recent call last):
            ...
            NotImplementedError: This functor can only be applied on the
            symbolic ring but Symbolic Subring accepting the variable a given.
        """
        if R is not SR:
            raise NotImplementedError('This functor can only be applied on '
                                      'the symbolic ring but %s given.' % (R,))
        return SymbolicSubring(accepting_variables=self.vars)


class SymbolicSubringRejectingVars(GenericSymbolicSubring):
    r"""
    The symbolic subring consisting of symbolic expressions whose variables
    are none of the given variables.
    """

    def _repr_(self):
        r"""
        Return a representation string of this symbolic subring.

        OUTPUT: string

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(rejecting_variables=('r',))  # indirect doctest
            Symbolic Subring rejecting the variable r
        """
        return 'Symbolic Subring rejecting %s' % \
            (self._repr_variables_())

    def has_valid_variable(self, variable):
        r"""
        Return whether the given ``variable`` is valid in this subring.

        INPUT:

        - ``variable`` -- a symbolic variable

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: S = SymbolicSubring(rejecting_variables=('r',))
            sage: S.has_valid_variable('a')
            True
            sage: S.has_valid_variable('r')
            False
            sage: S.has_valid_variable('x')
            True
        """
        return SR(variable) not in self._vars_

    def construction(self):
        r"""
        Return the functorial construction of this symbolic subring.

        OUTPUT:

        A tuple whose first entry is a construction functor and its second
        is the symbolic ring.

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(rejecting_variables=('r',)).construction()
            (Subring<rejecting r>, Symbolic Ring)
        """
        return (SymbolicSubringRejectingVarsFunctor(self._vars_), SR)

    def _coerce_map_from_(self, P):
        r"""
        Return whether ``P`` coerces into this symbolic subring.

        INPUT:

        - ``P`` -- a parent

        OUTPUT: boolean or ``None``

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: R = SymbolicSubring(rejecting_variables=('r',))
            sage: RS = SymbolicSubring(rejecting_variables=('r', 's'))
            sage: RS.has_coerce_map_from(R)  # indirect doctest
            False
            sage: R.has_coerce_map_from(RS)  # indirect doctest
            True
            sage: A = SymbolicSubring(accepting_variables=('a',))
            sage: R.has_coerce_map_from(A)
            True
            sage: A.has_coerce_map_from(R)
            False
        """
        if isinstance(P, SymbolicSubringRejectingVars):
            return self._vars_ <= P._vars_
        if isinstance(P, SymbolicSubringAcceptingVars):
            return not (self._vars_ & P._vars_)
        return super()._coerce_map_from_(P)

    def _an_element_(self):
        r"""
        Return an element of this symbolic subring.

        OUTPUT: a symbolic expression

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(rejecting_variables=('r',)).an_element()
            some_variable
            sage: _.parent()
            Symbolic Subring rejecting the variable r
            sage: SymbolicSubring(rejecting_variables=('some_variable',)).an_element()
            some_some_variable
            sage: _.parent()
            Symbolic Subring rejecting the variable some_variable
            sage: SymbolicSubring(rejecting_variables=('some_some_variable',)).an_element()
            some_variable
            sage: _.parent()
            Symbolic Subring rejecting the variable some_some_variable
            sage: SymbolicSubring(rejecting_variables=('some_variable','some_some_variable')).an_element()
            some_some_some_variable
            sage: _.parent()
            Symbolic Subring rejecting the variables some_some_variable, some_variable
        """
        v = SR.an_element()
        while not self.has_valid_variable(v):
            v = SR('some_' + str(v))
        return self(v)


class SymbolicSubringRejectingVarsFunctor(GenericSymbolicSubringFunctor):

    _functor_name = 'SymbolicSubringRejectingVarsFunctor'

    _repr_type_ = 'rejecting'

    def merge(self, other):
        r"""
        Merge this functor with ``other`` if possible.

        INPUT:

        - ``other`` -- a functor

        OUTPUT: a functor or ``None``

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F = SymbolicSubring(accepting_variables=('a',)).construction()[0]
            sage: G = SymbolicSubring(rejecting_variables=('r',)).construction()[0]
            sage: G.merge(G) is G
            True
            sage: G.merge(F) is G
            True
        """
        if self == other:
            return self
        if type(self) is type(other):
            return type(self)(self.vars & other.vars)
        if isinstance(other, SymbolicSubringAcceptingVarsFunctor):
            if not (self.vars & other.vars):
                return self

    def _apply_functor(self, R):
        """
        Apply this functor to the given symbolic ring `R`.

        INPUT:

        - ``R`` -- a symbolic ring

        OUTPUT: a subring of ``R``

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: F, R = SymbolicSubring(rejecting_variables=('r',)).construction()
            sage: F(R)  # indirect doctest
            Symbolic Subring rejecting the variable r

        TESTS::

            sage: F(F(R))
            Traceback (most recent call last):
            ...
            NotImplementedError: This functor can only be applied on the
            symbolic ring but Symbolic Subring rejecting the variable r given.
        """
        if R is not SR:
            raise NotImplementedError('This functor can only be applied on '
                                      'the symbolic ring but %s given.' % (R,))
        return SymbolicSubring(rejecting_variables=self.vars)


class SymbolicConstantsSubring(SymbolicSubringAcceptingVars):
    r"""
    The symbolic subring consisting of symbolic constants.
    """

    def _repr_(self):
        r"""
        Return a representation string of this symbolic subring.

        OUTPUT: string

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(no_variables=True)  # indirect doctest
            Symbolic Constants Subring
        """
        return 'Symbolic Constants Subring'

    def has_valid_variable(self, variable):
        r"""
        Return whether the given ``variable`` is valid in this subring.

        INPUT:

        - ``variable`` -- a symbolic variable

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: S = SymbolicSubring(no_variables=True)
            sage: S.has_valid_variable('a')
            False
            sage: S.has_valid_variable('r')
            False
            sage: S.has_valid_variable('x')
            False
        """
        return False

    def _an_element_(self):
        r"""
        Return an element of this symbolic subring.

        OUTPUT: a symbolic expression

        TESTS::

            sage: from sage.symbolic.subring import SymbolicSubring
            sage: SymbolicSubring(no_variables=True).an_element()
            I*pi*e
            sage: _.parent()
            Symbolic Constants Subring
        """
        return self(SR('I') * SR('pi') * SR('e'))
